/***************************************************************/
/*                                                             */
/* 2d pde problem in the unit square with linear finite        */
/* elements                                                    */
/*                                                             */
/***************************************************************/

#include "mpi.h"
#include "headers.h"

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>

#include "example_ansatz.h"


/* define the master process */
#define MASTER  0

/* start of main program */
int main(int argc, char *argv[]) 
{
  int rank, size;
  int x_subdoms,  y_subdoms, i, *info, x_int_loc, y_int_loc;
  int x_int_glob, y_int_glob, dof_glob, dof_loc;
  double t1, t2, h_x, h_y, x0_loc, y0_loc, x, y;
  int left_neighbour, right_neighbour, lower_neighbour, upper_neighbour;
  int *KCol, *RowPtr, j, start, end;
  int solver_type, maxite = 100000;
  double *matrix_entries, *rhs, *sol, *defect, res, damp, eps=1e-6;   
  
  /* initialize MPI */
  MPI_Init(&argc,&argv);

  /* every process asks its own number */
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  /* every process asks the total number of processes */
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  /* allocate integer array for data transfer */
  if (!(info=malloc(4*sizeof(*info))))
    MPI_Abort(MPI_COMM_WORLD, 1);
   
  /* the unit square will be divided in subdomains */
  /* the number of subdomains in x-direction will be read */
  /* the number of subdomains in y-direction will be computed */
  /* by number of processes/number of subdomains in x-direction */

  if (rank==MASTER) 
  {
    printf("number of subdomains in x-direction: (non-positive finishes the program): ");
    scanf("%d", &x_subdoms);
    if (x_subdoms<=0)
      /* kill all prozeses */
      MPI_Abort(MPI_COMM_WORLD, 1);

    /* compute number of subdomains in y-direction */
    y_subdoms = size/x_subdoms;

    if (y_subdoms<=0)
      /* kill all prozeses */
      MPI_Abort(MPI_COMM_WORLD, 1);

    // determine the grid size 
    printf("number of intervals in x-direction in each subdomain: ");
    printf("(non-positive finishes the program): ");
    scanf("%d", &x_int_loc);
    if (x_int_loc<=0)
      /* kill all prozeses */
      MPI_Abort(MPI_COMM_WORLD, 1);
    printf("number of intervals in y-direction in each subdomain: ");
    printf("(non-positive finishes the program): ");
    scanf("%d", &y_int_loc);
    if (y_int_loc<=0)
      /* kill all prozeses */
      MPI_Abort(MPI_COMM_WORLD, 1);

    /* fill array for transfering information to the other processes */
    info[0] = x_subdoms;
    info[1] = y_subdoms;
    info[2] = x_int_loc;
    info[3] = y_int_loc;
  }  
 
  /* braodcast information */
  MPI_Bcast(info, 4, MPI_INT, MASTER, MPI_COMM_WORLD);
  
  /* store information to the local variables */
  x_subdoms = info[0];
  y_subdoms = info[1];
  x_int_loc = info[2];
  y_int_loc = info[3];

  /***************************************************/
  /* stop if there are processors which are not      */
  /* involved in solving the problem (this makes the */
  /* coding easier if there are no such processors   */
  /***************************************************/
  if (rank >= x_subdoms * y_subdoms)
  {
     printf("number of subdomains %d does not match number of processors %d\n",
            x_subdoms * y_subdoms,size);
      MPI_Abort(MPI_COMM_WORLD, 1);
  }    
  /***************************************************/
  /* compute local and global parameters of the mesh */
  /***************************************************/
  /* number of local d.o.f. */
  dof_loc = (x_int_loc+1)*(y_int_loc+1);
  /* number of intervals in x-dirction (globally) */
  x_int_glob = x_int_loc*x_subdoms;
  /* number of intervals in y-dirction (globally) */
  y_int_glob = y_int_loc*y_subdoms;
  /* number of global d.o.f. */
  dof_glob = (x_int_glob+1)*(y_int_glob+1);
  if (rank==0)
  {
     printf("number of global degrees of freedom : %d ",dof_glob);
     printf("(inner d.o.f.: %d )\n ",(x_int_glob-1)*(y_int_glob-1));
  }
  /* mesh width in x-direction */
  h_x = 1.0/x_int_glob;
  /* mesh width in y-direction */
  h_y = 1.0/y_int_glob;
  
  /* compute neighbours */
  ComputeNeighbours(rank, x_subdoms, y_subdoms,
                    &left_neighbour, &right_neighbour,
                    &lower_neighbour, &upper_neighbour);
  
  /* coordinates of local left lower corner */
  LocalLeftLowerCorner(rank, x_subdoms, y_subdoms,&x0_loc,&y0_loc);
  
  /*************************************************/
  /* compute matrix structure, allocte arrays for  */
  /* solution and rhs                              */
  /*************************************************/
  
  /* allocate memory */
  /* not more than 7*dof_loc entries possible */
  if (! (RowPtr = malloc(dof_loc*sizeof(int)+1)))
     MPI_Abort(MPI_COMM_WORLD, 1);
  if (! (KCol = malloc(7*dof_loc*sizeof(int))))
     MPI_Abort(MPI_COMM_WORLD, 1);
  ComputeMatrixStructure(dof_loc, x_int_loc, y_int_loc, RowPtr,
                         KCol);
  
  /* allocate matrix entries and initialize */
  if (! (matrix_entries = malloc(7*dof_loc*sizeof(double))))
      MPI_Abort(MPI_COMM_WORLD, 1);
  memset(matrix_entries,0,7*dof_loc*sizeof(double));
  
  /* allocate right hand side and initialize */
  if (! (rhs = malloc(dof_loc*sizeof(double))))
     MPI_Abort(MPI_COMM_WORLD, 1);
  memset(rhs,0,dof_loc*sizeof(double));
  
  /* allocate solution and initialize */
  if (! (sol = malloc(dof_loc*sizeof(double))))
     MPI_Abort(MPI_COMM_WORLD, 1);
  memset(sol,0,dof_loc*sizeof(double));
  
  /**********************************/
  /* assembling of matrices and rhs */
  /**********************************/
  Assemble(dof_loc, x_int_loc, y_int_loc, x0_loc, y0_loc, x_subdoms, y_subdoms,
           h_x, h_y, RowPtr, KCol, matrix_entries, rhs);
  
  SetDirichletValues(left_neighbour, right_neighbour,
                     lower_neighbour, upper_neighbour,
                     x_int_loc, y_int_loc, x_subdoms, y_subdoms,
                     x0_loc , y0_loc, RowPtr, KCol, matrix_entries,
                     sol, rhs);
  if (rank==0)
  {
     for (i=0;i<dof_loc;i++)
     {
         printf("u(%d) =  %g ; \n ",i+1,sol[i]);       
     }     
  }
       
  /*********************/
  /* choose the solver */
  /*********************/
  
  if (rank==0)
  { 
     /* read solver type */
     printf("solver type:\n");
     printf("  1 - damped Jacobi \n");      
     printf("  2 - Block Jacobi with inner SOR \n");      
     printf("  3 - Block Jacobi with inner SSOR \n");      
     printf("  4 - Conjugate gradient \n");      
     printf("  5 - Preconditioned conjugate gradient with diagonal preconditioner \n");      
     scanf("%d", &solver_type);
     /*wrong solver type */
     if ((solver_type < 1)|| (solver_type>5))
     {
	printf("invalid solver type !!!\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
     }
  }
  /* send information to the other processors */
  MPI_Bcast(&solver_type, 1, MPI_INT, MASTER, MPI_COMM_WORLD);
  
  /* allocate defect */
  if (! (defect = malloc(dof_loc*sizeof(double))))
     MPI_Abort(MPI_COMM_WORLD, 1);
  memset(defect,0,dof_loc*sizeof(double));
  
  /* solve system */
  switch(solver_type)
  {
    case 1: /* damped Jacobi */
       /* read damping parameter */
       if (rank==0)
       { 	    
	  printf("damping parameter: ");
	  scanf("%lf", &damp);
       }
       MPI_Bcast(&damp, 1, MPI_DOUBLE, MASTER, MPI_COMM_WORLD);
       /* starting time */
       t1 = MPI_Wtime();
       /* call Jacobi iteration */
       jacobi(rank, left_neighbour, right_neighbour,
              lower_neighbour, upper_neighbour,
              x_int_loc, y_int_loc,
              maxite, eps, damp, dof_loc, KCol, RowPtr,
              matrix_entries, sol, rhs, defect);
       /* final time */
       t2 = MPI_Wtime();
       break;
     case 2: /* SOR */
       /* read damping parameter */
       if (rank==0)
       { 	    
	  printf("damping parameter: ");
	  scanf("%lf", &damp);
       }
       MPI_Bcast(&damp, 1, MPI_DOUBLE, MASTER, MPI_COMM_WORLD);
       /* starting time */
       t1 = MPI_Wtime();
       /* call SOR iteration */
       sor(rank, left_neighbour, right_neighbour,
           lower_neighbour, upper_neighbour,
           x_int_loc, y_int_loc,
           maxite, eps, damp, dof_loc, KCol, RowPtr,
           matrix_entries, sol, rhs, defect);
       /* final time */
       t2 = MPI_Wtime();
       break;
     case 3: /* SSOR */
       /* read damping parameter */
       if (rank==0)
       { 	    
	  printf("damping parameter: ");
	  scanf("%lf", &damp);
       }
       MPI_Bcast(&damp, 1, MPI_DOUBLE, MASTER, MPI_COMM_WORLD);
       /* starting time */
       t1 = MPI_Wtime();
       /* call SSOR iteration */
       ssor(rank, left_neighbour, right_neighbour,
            lower_neighbour, upper_neighbour,
            x_int_loc, y_int_loc,
            maxite, eps, damp, dof_loc, KCol, RowPtr,
            matrix_entries, sol, rhs, defect);
       /* final time */
       t2 = MPI_Wtime();
       break;
     case 4: /* CG */
       /* starting time */
       t1 = MPI_Wtime();
       /* call SSOR iteration */
       cg(rank, left_neighbour, right_neighbour,
          lower_neighbour, upper_neighbour,
          x_int_loc, y_int_loc,
          maxite, eps, damp, dof_loc, KCol, RowPtr,
          matrix_entries, sol, rhs, defect);
       /* final time */
       t2 = MPI_Wtime();
       break;
     case 5: /* PCG */
       /* starting time */
       t1 = MPI_Wtime();
       /* call SSOR iteration */
       pcg(rank, left_neighbour, right_neighbour,
           lower_neighbour, upper_neighbour,
           x_int_loc, y_int_loc,
           maxite, eps, damp, dof_loc, KCol, RowPtr,
           matrix_entries, sol, rhs, defect);
       /* final time */
       t2 = MPI_Wtime();
       break;
 }
/*
  if (rank==0)
  {
     for (i=0;i<dof_loc;i++)
     {
        start = RowPtr[i];
        end = RowPtr[i+1];
        printf("proc %d row %d : ",rank, i);
        for (j=start;j<end;j++)
        {
           printf("%d %g ",KCol[j],matrix_entries[j]);
        }
        printf("rhs %g\n",rhs[i]);
     }
}
*/
  if (rank==0)
  {
     for (i=0;i<dof_loc;i++)
     {
        start = RowPtr[i];
        end = RowPtr[i+1];
        for (j=start;j<end;j++)
        {
           printf("a(%d,%d) =  %g ; \n ",i+1,KCol[j]+1,matrix_entries[j]);
        }
        printf("f(%d) =  %g ; \n ",i+1,rhs[i]);       
     }     
}


/*
  for (i=0;i<dof_loc;i++)
  {
     ComputeCoordinates(x0_loc, y0_loc, i, x_subdoms, y_subdoms,  
                     x_int_loc, y_int_loc, &x, &y);
     sol[i] = Solution(x,y);
     }   
*/

  res=MatVecDefect(dof_loc,RowPtr, KCol, matrix_entries, sol, rhs, defect);


  for (i=0;i<dof_loc;i++)
  {
     ComputeCoordinates(x0_loc, y0_loc, i, x_subdoms, y_subdoms,  
                     x_int_loc, y_int_loc, &x, &y);
     
     /*printf("before %d %g %g %g \n",rank,x,y,defect[i]);*/
     }
  MakeVectorConsistent(left_neighbour, right_neighbour,
                       lower_neighbour, upper_neighbour,
                       x_int_loc, y_int_loc, defect);

/*  for (i=0;i<dof_loc;i++)
  {
     ComputeCoordinates(x0_loc, y0_loc, i, x_subdoms, y_subdoms,  
       x_int_loc, y_int_loc, &x, &y);
     
     printf("after %d %g %g %g \n",rank, x, y,defect[i]);
     }*/

  if (rank==MASTER)
     printf("solver time  %g \n",t2-t1);
  /* final MPI command */
  MPI_Finalize();

  return(EXIT_SUCCESS);
}


